const St          = imports.gi.St;
const Gio         = imports.gi.Gio
const Gtk         = imports.gi.Gtk;
const GLib        = imports.gi.GLib;
const Meta        = imports.gi.Meta;
const Shell       = imports.gi.Shell;
const Pango       = imports.gi.Pango;
const Clutter     = imports.gi.Clutter;
const Main        = imports.ui.main;
const CheckBox    = imports.ui.checkBox;
const PopupMenu   = imports.ui.popupMenu;
const MessageTray = imports.ui.messageTray;
const Slider      = imports.ui.slider;
const Lang        = imports.lang;
const Signals     = imports.signals;
const Mainloop    = imports.mainloop;


const ME = imports.misc.extensionUtils.getCurrentExtension();


const Gettext  = imports.gettext.domain(ME.metadata['gettext-domain']);
const _        = Gettext.gettext;
const ngettext = Gettext.ngettext;


const SOUND_PLAYER    = ME.imports.lib.sound_player;
const FULLSCREEN      = ME.imports.lib.fullscreen;
const SIG_MANAGER     = ME.imports.lib.signal_manager;
const KEY_MANAGER     = ME.imports.lib.keybinding_manager;
const PANEL_ITEM      = ME.imports.lib.panel_item;
const NUM_PICKER      = ME.imports.lib.num_picker;
const MULTIL_ENTRY    = ME.imports.lib.multiline_entry;
const TEXT_LINKS_MNGR = ME.imports.lib.text_links_manager;
const MISC_UTILS      = ME.imports.lib.misc_utils;
const REG             = ME.imports.lib.regex;


const IFACE = `${ME.path}/dbus/timer_iface.xml`;


const CACHE_FILE = GLib.get_home_dir() +
                   '/.cache/timepp_gnome_shell_extension/timepp_timer.json';


const TIMER_MAX_DURATION = 86400000000; // 24 hours in microseconds
const TIMER_EXPIRED_MSG  = _('Timer Expired!');


const TimerState = {
    RUNNING : 'RUNNING',
    STOPPED : 'STOPPED',
    OFF     : 'OFF',
};


const NotifStyle = {
    STANDARD   : 0,
    FULLSCREEN : 1,
};


// =====================================================================
// @@@ Main
//
// @ext      : obj (main extension object)
// @settings : obj (extension settings)
// =====================================================================
var SectionMain = new Lang.Class({
    Name    : 'Timepp.Timer',
    Extends : ME.imports.sections.section_base.SectionBase,

    _init: function (section_name, ext, settings) {
        this.parent(section_name, ext, settings);

        this.actor.add_style_class_name('timer-section');

        this.separate_menu = this.settings.get_boolean('timer-separate-menu');


        this.timer_state     = TimerState.OFF;
        this.tic_mainloop_id = null;
        this.cache_file      = null;
        this.cache           = null;
        this.notif_source    = null;
        this.clock           = 0; // microseconds
        this.end_time        = 0; // For computing elapsed time (microseconds)


        {
            let [,xml,] = Gio.file_new_for_path(IFACE).load_contents(null);
            xml = '' + xml;
            this.dbus_impl = Gio.DBusExportedObject.wrapJSObject(xml, this);
            this.dbus_impl.export(Gio.DBus.session, '/timepp/zagortenay333/Timer');
        }


        this.linkm = new TEXT_LINKS_MNGR.TextLinksManager(MISC_UTILS.split_on_whitespace);
        this.sigm  = new SIG_MANAGER.SignalManager();
        this.keym  = new KEY_MANAGER.KeybindingManager(this.settings);


        this.sound_player = new SOUND_PLAYER.SoundPlayer();


        this.fullscreen = new TimerFullscreen(this.ext, this,
            this.settings.get_int('timer-fullscreen-monitor-pos'));
        this.fullscreen.set_banner_text(
            this.settings.get_boolean('timer-show-seconds') ? '00:00:00' : '00:00');


        try {
            this.cache_file = Gio.file_new_for_path(CACHE_FILE);

            let cache_format_version =
                ME.metadata['cache-file-format-version'].timer;

            if (this.cache_file.query_exists(null)) {
                let [a, contents, b] = this.cache_file.load_contents(null);
                this.cache = JSON.parse(contents);
            }

            if (!this.cache || !this.cache.format_version ||
                this.cache.format_version !== cache_format_version) {

                this.cache = {
                    format_version         : cache_format_version,
                    notif_msg              : '',
                    last_manually_set_time : 30, // seconds
                };
            }
        }
        catch (e) {
            logError(e);
            return;
        }


        //
        // keybindings
        //
        this.keym.register('timer-keybinding-open', () => {
             this.ext.open_menu(this.section_name);
        });
        this.keym.register('timer-keybinding-open-fullscreen', () => {
            this.show_fullscreen();
        });


        //
        // panel item
        //
        this.panel_item.icon.icon_name = 'timepp-timer-symbolic';
        this.panel_item.actor.add_style_class_name('timer-panel-item');
        this.panel_item.set_label(this.settings.get_boolean('timer-show-seconds') ? '00:00:00' : '00:00');
        this._toggle_panel_item_mode();


        //
        // item with the time display, switcher and settings icon
        //
        this.header = new PopupMenu.PopupMenuItem(_('Timer'), { hover: false, activate: false, style_class: 'header' });
        this.header.actor.can_focus = false;
        this.header.label.add_style_class_name('clock');
        this.actor.add_actor(this.header.actor);

        this.icon_box = new St.BoxLayout({ y_align: Clutter.ActorAlign.CENTER, x_align: Clutter.ActorAlign.END, style_class: 'icon-box' });
        this.header.actor.add(this.icon_box, {expand: true});

        this.toggle = new PopupMenu.Switch('');
        this.toggle_bin = new St.Button({ visible: false, can_focus: true, y_align: St.Align.MIDDLE });
        this.toggle_bin.add_actor(this.toggle.actor);
        this.icon_box.add(this.toggle_bin);

        this.fullscreen_bin  = new St.Button({ can_focus: true, y_align: St.Align.MIDDLE, x_align: St.Align.END, style_class: 'fullscreen-icon' });
        this.icon_box.add(this.fullscreen_bin);
        this.fullscreen_icon = new St.Icon({ icon_name: 'timepp-fullscreen-symbolic' });
        this.fullscreen_bin.add_actor(this.fullscreen_icon);

        this.settings_icon = new St.Icon({ icon_name: 'timepp-settings-symbolic' });
        this.settings_bin  = new St.Button({ can_focus: true, y_align: St.Align.MIDDLE, x_align: St.Align.END, style_class: 'settings-icon' });
        this.settings_bin.add_actor(this.settings_icon);
        this.icon_box.add(this.settings_bin);


        //
        // timer slider
        //
        this.slider_item = new PopupMenu.PopupBaseMenuItem({ activate: false });
        this.actor.add_actor(this.slider_item.actor);

        this.slider = new Slider.Slider(0);
        this.slider_item.actor.add(this.slider.actor, { expand: true });


        //
        // settings window container
        //
        this.timepicker_container = new St.Bin({ x_fill: true });
        this.actor.add_child(this.timepicker_container);


        //
        // listen
        //
        this.sigm.connect(this.fullscreen, 'monitor-changed', () => {
            this.settings.set_int('timer-fullscreen-monitor-pos', this.fullscreen.monitor);
        });
        this.sigm.connect(this.settings, 'changed::timer-separate-menu', () => {
            this.separate_menu = this.settings.get_boolean('timer-separate-menu');
            this.ext.update_panel_items();
        });
        this.sigm.connect(this.settings, 'changed::timer-show-seconds', () => this._update_time_display());
        this.sigm.connect(this.settings, 'changed::timer-panel-mode', () => this._toggle_panel_item_mode());
        this.sigm.connect(this.panel_item, 'left-click', () => this.ext.toggle_menu(this.section_name));
        this.sigm.connect(this.panel_item, 'middle-click', () => this.toggle_timer());
        this.sigm.connect_press(this.toggle_bin, Clutter.BUTTON_PRIMARY, true, () => this.toggle_timer());
        this.sigm.connect_press(this.fullscreen_bin, Clutter.BUTTON_PRIMARY, true, () => this.show_fullscreen());
        this.sigm.connect_press(this.settings_bin, Clutter.BUTTON_PRIMARY, true, () => this._show_settings());
        this.sigm.connect(this.slider, 'value-changed', (slider, value) => this.slider_changed(slider, value));
        this.sigm.connect(this.slider, 'drag-end', () => this.slider_released());
        this.sigm.connect(this.slider.actor, 'scroll-event', () => this.slider_released());
        this.sigm.connect(this.slider_item.actor, 'button-press-event', (_, event) => this.slider.startDragging(event));
    },

    disable_section: function () {
        this.dbus_impl.unexport();
        this.stop();
        this._store_cache();
        this.sigm.disconnect_all();
        this.keym.disable_all();

        if (this.fullscreen) {
            this.fullscreen.destroy();
            this.fullscreen = null;
        }

        this.parent();
    },

    _store_cache: function () {
        if (! this.cache_file.query_exists(null))
            this.cache_file.create(Gio.FileCreateFlags.NONE, null);

        this.cache_file.replace_contents(JSON.stringify(this.cache, null, 2),
            null, false, Gio.FileCreateFlags.REPLACE_DESTINATION, null);
    },

    toggle_timer: function () {
        if      (this.timer_state === TimerState.STOPPED) this.start();
        else if (this.timer_state === TimerState.RUNNING) this.stop();
    },

    // @time: int (seconds)
    start: function (time) {
        if (this.tic_mainloop_id) {
            Mainloop.source_remove(this.tic_mainloop_id);
            this.tic_mainloop_id = null;
        }

        this.sound_player.stop();
        if (this.notif_source) this.notif_source.destroyNonResidentNotifications();

        this.timer_state = TimerState.RUNNING;

        if (time) time *= 1000000;
        else      time  = this.clock;

        this.end_time = GLib.get_monotonic_time() + time;

        this.fullscreen.on_timer_started();
        this.toggle.setToggleState('checked');
        this.toggle_bin.show();
        this.panel_item.actor.add_style_class_name('on');

        this._tic();
    },

    stop: function () {
        this.clock = this.end_time - GLib.get_monotonic_time();

        if (this.tic_mainloop_id) {
            Mainloop.source_remove(this.tic_mainloop_id);
            this.tic_mainloop_id = null;
        }

        this.fullscreen.on_timer_stopped();
        this.timer_state = TimerState.STOPPED;
        this.toggle.setToggleState('');
        this.panel_item.actor.remove_style_class_name('on');
    },

    reset: function () {
        if (this.tic_mainloop_id) {
            Mainloop.source_remove(this.tic_mainloop_id);
            this.tic_mainloop_id = null;
        }

        this.slider.setValue(0);
        this.fullscreen.on_timer_off();
        this.timer_state = TimerState.OFF;
        this.header.label.text = _('Timer');
        this.toggle_bin.hide();
        this.panel_item.actor.remove_style_class_name('on');
    },

    _on_timer_expired: function () {
        this.reset();
        this.fullscreen.on_timer_expired();
        this._send_notif();
        this.dbus_impl.emit_signal('timer_expired', null);
    },

    _tic: function () {
        this.clock = this.end_time - GLib.get_monotonic_time();

        if (this.clock <= 0) {
            this.clock = 0;
            this._on_timer_expired();
            return;
        }

        this._update_slider();
        this._update_time_display();

        this.tic_mainloop_id = Mainloop.timeout_add_seconds(1, () => {
            this._tic();
        });
    },

    _update_time_display: function () {
        let time = Math.ceil(this.clock / 1000000);
        let txt;

        // If the seconds are not shown, we need to make the timer '1-indexed'
        // in respect to minutes. I.e., 00:00:34 becomes 00:01.
        if (this.settings.get_boolean('timer-show-seconds')) {
            txt = "%02d:%02d:%02d".format(
                Math.floor(time / 3600),
                Math.floor(time % 3600 / 60),
                time % 60
            );
        }
        else {
            if (time % 3600 !== 0) time += 60;

            txt = "%02d:%02d".format(
                Math.floor(time / 3600),
                Math.floor(time % 3600 / 60)
            );
        }

        this.header.label.text = txt;
        this.panel_item.set_label(txt);
        this.fullscreen.set_banner_text(txt);
    },

    // Update slider based on the clock.
    // This function is the inverse of the function that is used to calc the
    // clock based on the slider.
    _update_slider: function () {
        let x = this.clock / TIMER_MAX_DURATION;
        let y = (Math.log(x * (Math.pow(2, 10) - 1) +1)) / Math.log(2) / 10;

        this.slider.setValue(y);
        this.fullscreen.slider.setValue(y);
    },

    slider_released: function () {
        if (this.clock < 1000000) {
            this.reset();
        }
        else {
            this.start();
            this.cache.last_manually_set_time = Math.round(this.clock / 1000000);
            this._store_cache();
        }
    },

    slider_changed: function (slider, value) {
        this.stop();

        if (value < 1) {
            // Make rate of change of the timer duration an exponential curve.
            // This allows for finer tuning when the duration is smaller.
            let y = (Math.pow(2, (10 * value)) - 1) / (Math.pow(2, 10) - 1);

            // Change the increment of the slider based on how far it's dragged.
            // If the seconds are not shown, the increments must be multiples
            // of 60s.
            let step;

            if (this.settings.get_boolean('timer-show-seconds')) {
                if      (value < .05) step = 15;
                else if (value < .5)  step = 30;
                else if (value < .8)  step = 60;
                else                  step = 3600;
            }
            else {
                if      (value < .7)  step = 59;
                else if (value < .9)  step = 1800;
                else                  step = 3600;
            }

            this.clock = Math.floor(y * TIMER_MAX_DURATION / step) * step;

            this._update_time_display();
        }
        else { // slider has been dragged past the limit
            this.clock = TIMER_MAX_DURATION;
            this._update_time_display();
        }
    },

    _send_notif: function () {
        if (this.settings.get_boolean('timer-play-sound')) {
            this.sound_player.set_sound_uri(this.settings.get_string('timer-sound-file-path'));
            this.sound_player.play(this.settings.get_boolean('timer-do-repeat-notif-sound'));
        }

        if (this.fullscreen.is_open) {
            return;
        }

        if (this.settings.get_enum('timer-notif-style') === NotifStyle.FULLSCREEN) {
            this.fullscreen.open();
            return;
        }

        if (this.notif_source) {
            this.notif_source.destroyNonResidentNotifications();
        }

        this.notif_source = new MessageTray.Source();
        Main.messageTray.add(this.notif_source);
        this.notif_source.connect('destroy', () => this.sound_player.stop());

        let icon = new St.Icon({ icon_name: 'timepp-timer-symbolic' });

        let params = {
            bannerMarkup : true,
            gicon        : icon.gicon,
        };

        let notif = new MessageTray.Notification(
            this.notif_source,
            TIMER_EXPIRED_MSG,
            this.cache.notif_msg,
            params
        );

        notif.setUrgency(MessageTray.Urgency.CRITICAL);

        this.notif_source.notify(notif);
    },

    _show_settings: function () {
        let settings = new TimerSettings(
            this.ext,
            this,
            this.settings.get_boolean('timer-show-seconds'),
            this.cache.notif_msg
        );

        this.timepicker_container.add_actor(settings.actor);
        settings.button_cancel.grab_key_focus();

        this.header.actor.hide();
        this.slider_item.actor.hide();

        settings.connect('ok', (_, info) => {
            this.actor.grab_key_focus();
            settings.actor.destroy();
            this.header.actor.show();
            this.slider_item.actor.show();

            this.settings.set_boolean('timer-do-repeat-notif-sound', info.repeat_sound);
            this.set_notif_msg(info.notif_msg);

            if (info.time) {
                this.clock = info.time;
                this.start();
                this._update_slider();
                this.cache.last_manually_set_time = info.time;
                this._store_cache();
            }
        });

        settings.connect('cancel', () => {
            this.actor.grab_key_focus();
            settings.actor.destroy();
            this.header.actor.show();
            this.slider_item.actor.show();
        });
    },

    set_notif_msg: function (msg) {
        this.cache.notif_msg = msg;
        this._store_cache();
    },

    show_fullscreen: function () {
        this.ext.menu.close();

        if (! this.fullscreen) {
            this.fullscreen = new TimerFullscreen(
                this.ext, this, this.settings.get_int('timer-fullscreen-monitor-pos'));
        }

        this.fullscreen.open();
    },

    _toggle_panel_item_mode: function () {
        if (this.settings.get_enum('timer-panel-mode') === 0)
            this.panel_item.set_mode('icon');
        else if (this.settings.get_enum('timer-panel-mode') === 1)
            this.panel_item.set_mode('text');
        else
            this.panel_item.set_mode('icon_text');
    },
});
Signals.addSignalMethods(SectionMain.prototype);



// =====================================================================
// @@@ Settings window
//
// @ext       : obj (main extension object)
// @delegate  : obj (main section object)
// @show_secs : bool
// @notif_msg : string
//
// signals: 'ok', 'cancel'
// =====================================================================
const TimerSettings = new Lang.Class({
    Name: 'Timepp.TimerSettings',

    _init: function(ext, delegate, show_secs, notif_msg) {
        this.ext      = ext;
        this.delegate = delegate;

        this.actor = new St.Bin({ x_fill: true, style_class: 'view-box' });

        this.content_box = new St.BoxLayout({ x_expand: true, vertical: true, style_class: 'view-box-content' });
        this.actor.add_actor(this.content_box);


        //
        // time pickers
        //
        {
            let box = new St.BoxLayout({ style_class: 'row numpicker-box' });
            this.content_box.add_actor(box);

            let label = new St.Label({ x_expand: true, y_align: Clutter.ActorAlign.CENTER });
            box.add_child(label);

            this.hr = new NUM_PICKER.NumPicker(0, 23);
            box.add_child(this.hr.actor);

            this.min = new NUM_PICKER.NumPicker(0, 59);
            box.add_child(this.min.actor);

            if (show_secs) {
                label.text = `${_('(h:min:sec)')} `;

                this.sec = new NUM_PICKER.NumPicker(0, 59);
                box.add_child(this.sec.actor);
            }
            else {
                label.text = `${_('(h:min)')} `;
            }
        }


        //
        // entry
        //
        this.entry_container = new St.BoxLayout({ x_expand: true, style_class: 'row entry-container' });
        this.content_box.add_actor(this.entry_container);

        this.entry = new MULTIL_ENTRY.MultiLineEntry(_('Timer message...'), true);
        this.entry_container.add(this.entry.actor, {expand: true});

        this.entry.scroll_box.vscrollbar_policy = Gtk.PolicyType.NEVER;
        this.entry.scroll_box.hscrollbar_policy = Gtk.PolicyType.NEVER;

        this.entry.set_text(notif_msg);


        //
        // repeat sound checkbox
        //
        this.checkbox_item = new St.BoxLayout({ reactive: true, x_expand: true, style_class: 'row' });
        this.content_box.add_actor(this.checkbox_item);

        this.checkbox_item.add_child(
            new St.Label({ text: _('Repeat notification sound?'), x_expand: true, y_align: Clutter.ActorAlign.CENTER }));

        this.sound_checkbox = new CheckBox.CheckBox();
        this.checkbox_item.add_child(this.sound_checkbox.actor);
        this.sound_checkbox.actor.checked = this.delegate.settings.get_boolean('timer-do-repeat-notif-sound');


        //
        // buttons
        //
        let btn_box = new St.BoxLayout({ style_class: 'row btn-box' });
        this.content_box.add(btn_box, {expand: true});

        this.button_cancel = new St.Button({ can_focus: true, label: _('Cancel'), style_class: 'button', x_expand: true });
        this.button_ok     = new St.Button({ can_focus: true, label: _('Ok'), style_class: 'button', x_expand: true });
        btn_box.add(this.button_cancel, {expand: true});
        btn_box.add(this.button_ok, {expand: true});


        //
        // listen
        //
        this.entry.entry.connect('queue-redraw', () => {
            this.entry.scroll_box.vscrollbar_policy = Gtk.PolicyType.NEVER;

            if (ext.needs_scrollbar())
                this.entry.scroll_box.vscrollbar_policy = Gtk.PolicyType.ALWAYS;
        });
        this.button_ok.connect('clicked', () => {
            this.emit('ok', {
                time         : this._get_time(),
                msg          : this.entry.entry.get_text(),
                repeat_sound : this.sound_checkbox.actor.checked,
            });
        });
        this.button_cancel.connect('clicked', () => this.emit('cancel'));
        this.checkbox_item.connect('button-press-event', () => {
            this.sound_checkbox.actor.checked = !this.sound_checkbox.actor.checked;
        });
    },

    _get_time: function () {
        let h   = this.hr.counter * 3600;
        let min = this.min.counter * 60;
        let sec = this.sec ? this.sec.counter : 0;

        return (h + min + sec) * 1000000;
    },
});
Signals.addSignalMethods(TimerSettings.prototype);



// =====================================================================
// @@@ Timer fullscreen interface
//
// @ext      : obj (main extension object)
// @delegate : obj (main section object)
// @monitor  : int
//
// signals: 'monitor-changed'
// =====================================================================
const TimerFullscreen = new Lang.Class({
    Name    : 'Timepp.TimerFullscreen',
    Extends : FULLSCREEN.Fullscreen,

    _init: function (ext, delegate, monitor) {
        this.parent(monitor);
        this.default_style_class = this.actor.style_class;

        this.ext      = ext;
        this.delegate = delegate;

        this.css = this.ext.custom_css;

        this.delegate.linkm.add_label_actor(this.banner, new Map([
            [REG.URL       , MISC_UTILS.open_web_uri],
            [REG.FILE_PATH , MISC_UTILS.open_file_path],
        ]));


        //
        // actors
        //
        this.title = new St.Label({ x_expand: true, x_align: Clutter.ActorAlign.CENTER, style_class: 'pomo-phase-label' });
        this.middle_box.insert_child_at_index(this.title, 0);


        this.slider = new Slider.Slider(0);
        this.bottom_box.add_child(this.slider.actor);
        this.slider.actor.can_focus = true;


        this.toggle_bin = new St.Button({ can_focus: true, y_align: St.Align.MIDDLE });
        this.top_box.insert_child_at_index(this.toggle_bin, 0);
        this.toggle_bin.hide();
        this.toggle = new PopupMenu.Switch('');
        this.toggle_bin.add_actor(this.toggle.actor);


        //
        // listen
        //
        this.toggle_bin.connect('clicked', () => {
            this.delegate.toggle_timer();
        });
        this.slider.connect('drag-end', () => {
            this.delegate.slider_released();
        });
        this.slider.actor.connect('scroll-event', () => {
            this.delegate.slider_released();
            this.title.text = '';
        });
        this.slider.connect('value-changed', (slider, val) => {
            this.delegate.slider_changed(slider, val);
            this.actor.remove_style_class_name('timer-expired');
            this.title.text = '';
        });
        this.actor.connect('key-release-event', (_, event) => {
            switch (event.get_key_symbol()) {
                case Clutter.KEY_space:
                    if (this.delegate.timer_state !== TimerState.OFF)
                        this.delegate.toggle_timer();
                    return Clutter.EVENT_STOP;
                case Clutter.KEY_r:
                case Clutter.KEY_BackSpace:
                    this.delegate.start(this.delegate.cache.last_manually_set_time);
                    return Clutter.EVENT_STOP;
                case Clutter.KEY_1:
                    this.delegate.start(60);
                    return Clutter.EVENT_STOP;
                case Clutter.KEY_2:
                    this.delegate.start(2 * 60);
                    return Clutter.EVENT_STOP;
                case Clutter.KEY_3:
                    this.delegate.start(3 * 60);
                    return Clutter.EVENT_STOP;
                case Clutter.KEY_4:
                    this.delegate.start(4 * 60);
                    return Clutter.EVENT_STOP;
                case Clutter.KEY_5:
                    this.delegate.start(5 * 60);
                    return Clutter.EVENT_STOP;
                case Clutter.KEY_6:
                    this.delegate.start(6 * 60);
                    return Clutter.EVENT_STOP;
                case Clutter.KEY_7:
                    this.delegate.start(7 * 60);
                    return Clutter.EVENT_STOP;
                case Clutter.KEY_8:
                    this.delegate.start(8 * 60);
                    return Clutter.EVENT_STOP;
                case Clutter.KEY_9:
                    this.delegate.start(9 * 60);
                    return Clutter.EVENT_STOP;
                case Clutter.KEY_0:
                    this.delegate.start(10 * 60);
                    return Clutter.EVENT_STOP;
                default:
                    return Clutter.EVENT_PROPAGATE;
            }
        });
    },

    close: function () {
        this.delegate.sound_player.stop();

        if (this.delegate.timer_state === TimerState.OFF) {
            this.actor.style_class = this.default_style_class;
            this.title.text = '';
            this.set_banner_text(
                this.delegate.settings.get_boolean('timer-show-seconds') ? '00:00:00' : '00:00');
        }

        this.parent();
    },

    on_timer_started: function () {
        this.actor.style_class = this.default_style_class;
        this.title.text = '';
        this.toggle.setToggleState('checked');
        this.toggle_bin.show();
    },

    on_timer_stopped: function () {
        this.actor.style_class = this.default_style_class + ' timer-stopped';
        this.toggle.setToggleState('');
    },

    on_timer_off: function () {
        this.toggle_bin.hide();
        this.slider.setValue(0);
    },

    on_timer_expired: function () {
        if (this.delegate.cache.notif_msg) {
            this.title.text = TIMER_EXPIRED_MSG;

            this.set_banner_text(
                this._highlight_tokens(this.delegate.cache.notif_msg)
                    .replace(/&(?!amp;|quot;|apos;|lt;|gt;)/g, '&amp;')
                    .replace(/<(?!\/?[^<]*>)/g, '&lt;')
            );
        }
        else {
            this.set_banner_text(TIMER_EXPIRED_MSG);
        }
        this.actor.style_class = this.default_style_class + ' timer-expired';
    },

    _highlight_tokens: function (text) {
        text = MISC_UTILS.split_on_whitespace(
            MISC_UTILS.markup_to_pango(text, this.ext.markup_map));

        let token;

        for (let i = 0; i < text.length; i++) {
            token = text[i];

            if (REG.URL.test(token) || REG.FILE_PATH.test(token)) {
                text[i] =
                    '<span foreground="' + this.css['-timepp-link-color'][0] +
                    '"><u><b>' + token + '</b></u></span>';
            }
        }

        return text.join(' ').replace(/ \n /g, '\n');
    },
});
Signals.addSignalMethods(TimerFullscreen.prototype);
